require "Actiontables"
require "B_syntaxbume"
require "scanGeneric"

##
# Author:: Bauer, Sebastian
#
#	Parser (08.03.2006)
# ===================
#
# Dieses Programm kann anhand eines Automaten einen Parser erstellen.
# Dieser kann zu einem gegebenen Ausdruck die Reduktionsschritte 
# oder auch den Syntaxbaum bestimmen.

##
# Diese Klasse reprsentiert den Parser.
#
# Ihre Attribute sind:
# - @atable -> Die Aktionstabelle aus dem Automaten
# - @utable -> Die bergangstabelle aus dem Automaten
# - @start -> Der Startzustand des Automaten in den Tabellen
# - @symbols -> Alle Symbole als Strings in einem Array
class Parser
	
	##
	# Konstruktor von Parser:
	# Berechnet aus dem Automaten Aktions- und bergangstabelle, den Startzustans
	# und die Symbole. Diese werden in den jeweiligen Attributen abgelegt.
	#
	# Parameter:
	# - dea -> Ein deterministischer Automat (LR_0, LR_1, SLR_1)
	def initialize(dea)
		@atable = ATable.new(dea)
		@utable = UTable.new(dea)
		@start = @utable.getnamehs.invert[dea.getStartZustand().to_s]
		@symbols = @utable.getsymbole.values
	end
	
	##
	# Prft einen Ausdruck auf Korrektheit im Parser. Sammelt alle Reduktionsschritte
	# und liefert diese zurck.
	#
	# Parameter:
	# - str -> Der Ausdruck der geparst werden soll (Leerzeichen sind nicht ntig!)
	#
	# Return:
	# -> Ein Array mit Reduktionsitem-Strings (letzte Reduktion zuerst)
	# -> nil fals der Ausdruck nicht akzeptiert wird
	def parse(str)
		zust = @start
		left = Array.new
		right = Array.new
		regs = Array.new
		# Fr alle Symbole den passenden regulren Ausdruck in regs ablegen
		@symbols.each {|symb|
			if symb == "." or  symb == "|" or symb == "(" or symb ==")" or symb == "[" or symb == "{" or symb == "+" or symb == "\\" or symb == "^" or symb == "$" or symb == "*" or symb == "?"
				symb = "\\" + symb
			end
			regs.push(Regexp.new(symb))
		}
		scanner = Scanner.new([@symbols , regs])
		begin
			middle = scanner.zerlege(str)
		rescue
			return nil
		end	
		middle.push("#")
		progress = true
		# Nach Initialisierung:
		# zust = Startzustand
		# left, right = leer
		# middle = Array aller gescannten Symbole aus dem Ausdruck
		# progress = true
		#
		# Schleife fhrt die Aktionen der Aktionstabelle @atable aus und berechnet fr 
		# jeden Durchlauf den aktuellen Zustand(zust), die eingelesenen Symbole(left),
		# die noch einzulesenden Symbole(middle) und die bereits durchgefhrten 
		# Reduktionen(right) solange bis der Ausdruck komplett durchgelaufen ist oder
		# er nicht zur Grammatik passt.
		while progress
			# Prft ob der aktuelle Zustand bzw. die Zeile der Aktionstabelle leer ist 
			# (knnte bei nicht akzeptablen Ausdruck auftreten)
			if @atable.gettabelle[zust] != nil and zust != nil
				action = @atable.gettabelle[zust][middle[0]]
				# In der Aktionstabelle wird die Aktion gesucht die in der Zeile zust in der Spalte
				# des ersten Symbols aus middle steht. Abhngig von der dort stehenden Aktion wird
				# weiter verfahren.
				case action
					# Wenn in der Tabelle an der angegebenen Stelle "error" steht, kann das nchste Symbol
					# nach der Grammatik nicht als nchstes auftreten. Es wird abgebrochen und nil
					# zurckgegeben.
					when "error"
						return nil
					# Wenn in der Tabelle an der angegebenen Stelle "accept" steht, wurde der Komplette Ausdruck
					# akzeptiert und die Reduktionsschritte gesammelt. Die Schleife kann verlassen werden
					# (progress = false)
					when "accept"
						progress = false
					# Wenn in der Tabelle an der angegebenen Stele "shift" steht, wird das nchste einzulesende
					# Symbol gelesen (das erste Element von middle wird als letztes Element von left angefhrt und
					# aus middle gelscht) Der neue Zustand ergibt sich aus aktueller Zustand / eingelesenes Symbol
					# in der bergangstabelle
					when "shift"
						left.push(middle[0])
						zust = @utable.gettabelle[zust][middle[0]]
						middle.slice!(0)
					# Wenn keiner der drei oberen Flle aufgetreten ist, kann nur noch eine Reduktion an der
					# angegebenen Stelle angegeben sein. Der Tabelleneintrag wird zu right hinzugefgt. 
					# Anhand der Reduktionsregel werden die eingelsenen Symbole 
					else
						right.push(action)
						links,rechts = action.split "->"
						links = links.gsub("red[" , "").strip
						rechts = rechts.gsub(".]", "").strip
						rArray = scanner.zerlege(rechts)
						# Lsche jedes gescannte Symbol der rechten Reduktionsseite von links aus left
						# wenn sie bereinstimmen. Ansonsten wird abgebrochenund nil zurckgegeben
						rArray.reverse_each { |symb|
								(symb == left.last) ? (left.delete_at((left.length) -1)) : (return nil)
							}
							left.push(links)
							zust = @start
							# Nach der Reduktion muss der neue Zustand gefunden werden. Dafr wird zust
							# mit dem Startzustand initialisiert und fr jedes Symbol von left werden 
							# anhand der bergangstabelle die Zustnde weitergeschaltet
							left.each { |symb|
								zust = @utable.gettabelle[zust][symb]
							}
					end
			# Bricht ab fals der Zustand = nil ist und gibt nil zurck
			else
				return nil
			end
		end
		right.reverse
	end
	
	##
	# Erzeugt zu einem Ausdruck den Syntaxbaum. Dabei wird das Ergebnis von parse(str) genutzt
	#
	# Parameter:
	# - str -> Der Ausdruck der geparst und als Syntaxbaum zurckgegeben werden soll
	#
	#	Return:
	# -> Syntaxbaum der beginnend mit dem Startsymbol die Ableitungen bis hin zum Ausdruck enthlt
	# -> Nil falls der Ausdruck nicht geparst werden kann
	def to_SB (str)
		regs = Array.new
		# Fr alle Symbole den passenden regulren Ausdruck in regs ablegen
		@symbols.each {|symb|
			if symb == "." or  symb == "|" or symb == "(" or symb ==")" or symb == "[" or symb == "{" or symb == "+" or symb == "\\" or symb == "^" or symb == "$" or symb == "*" or symb == "?"
				symb = "\\" + symb
			end
			regs.push(Regexp.new(symb))
		}
		scanner = Scanner.new([@symbols , regs])
		redlist = parse(str)
		(redlist == nil) ? (return nil) : ()
		tree = Syntaxbaum.new("@",0)
		# Der Wurzelknoten wird immer mit "@" initialisiert. Danach wird jede Reduktion nacheinander
		# zerlegt und an den entsprechenden Stellen im Baum die passenden Symbole hinzugefgt.
		redlist.each { |red|
			links,rechts = red.split "->"
			links = links.gsub("red[" , "").strip
			rechts = rechts.gsub(".]", "").strip
			rechtsarr = scanner.zerlege(rechts)
			# Fr die erste Reduktion kann ohne weiteres am Wurzelknoten begonnen werden
			if tree.getSoehne.size == 0
				tree.addNonterminal(links,0)
				# Fgt jedes Symbol der rechten Regelseite an den Knoten an.
				rechtsarr.each {|symb|
					tree.getSoehne[0].addNonterminal(symb, tree.getSoehne[0].getSoehne.size)
				}
			# Fr jede weitere ausser der ersten Reduktion, muss erst der passende Knoten
			# im Baum gefunden werden um an diesen die Symbole anzufgen.
			else
				found = search(tree, links)
				# Fgt jedes Symbol der rechten Regelseite an den Knoten an.
				rechtsarr.each {|symb|
					found.addNonterminal(symb, found.getSoehne.size)
				}
			end
		}
		tree
	end
	
	##
	# Sucht in einem Teilbaum rekursiv den ersten Knoten von Rechts dessen 
	# Name mit einem bestimmten String bereinstimmt und keine Soehne hat.
	#
	# Parameter:
	# - knoten -> Der Wurzelknoten des Teilbaums
	# - symb -> Der Name des gesuchten Knotens
	#
	# Return: 
	# -> Der erste Knoten von rechts der die Bedingungen erfllt
	# -> nil (nur in der Rekursion) wenn der Knoten zwar keine Shne hat aber 
	#    der Name nicht mit dem Symbol bereinstimmt
	def search (knoten, symb)
		if knoten.getSoehne.size == 0
			if knoten.getName == symb
				# Wenn der Knoten keine Shne hat und der Name stimmt wird der Knoten zurckgegeben.
				# Er ist der gesuchte Knoten.
				return knoten
			else
				# Wenn der Knoten zwar keine Shne hat aber der Name nicht stimmt wird nil zurckgegeben
				#  Es ist nicht der gesuchte Knoten und es gibt keine Shne zum weitersuchen
				return nil
			end
		# Wenn der Knoten Shne hat kann er nicht der gesuchte Knoten sein. Es wird
		# mit den Shnen als Wurzelknoten neuer Teilbume weitergesucht.
		else
			# Sucht jeden Sohn von rechts nach links nach dem gesuchten Knoten ab.
			# sobald er gefunden wird er gefunden wird die Suche beendet und der Knoten
			# wird zurckgegeben. Wenn kein Sohn den gesuchten Knoten enthlt wird nil zurckgegeben.
			knoten.getSoehne.reverse.each {|sohn|
				a = search(sohn,symb)
				(a != nil) ? (return a) : ()
			}
			return nil
		end
	end
end